/*  horizon, a sketch, note and scribbling application.
    Copyright (C) 2005 Øyvind Kolås <pippin@gimp.org>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
*/
#include <glib.h>
#include <glib/gprintf.h>
#include <glib/gstdio.h>

#include <string.h>  /* memset */
#include "canvas.h"
#include "map.h"
#include "tile.h"
#include "tile_cache.h"

#include "config.h"

static gchar hex_digits[16]="0123456789ABCDEF";

static GObjectClass *parent_class = NULL;

static void
canvas_finalize (GObject *object)
{
  Canvas *canvas = (Canvas*) object;

  g_object_unref (G_OBJECT (canvas->cache));
  g_object_unref (G_OBJECT (canvas->system_cache));

  if (canvas->path)
    g_free (canvas->path);

  (* G_OBJECT_CLASS (parent_class)->finalize) (object);
}

static void
canvas_class_init (CanvasClass *class)
{
  GObjectClass *gobject_class;

  gobject_class = (GObjectClass*) class;

  parent_class = g_type_class_peek_parent (class);

  gobject_class->finalize = canvas_finalize;
}

static void
canvas_init (Canvas *canvas)
{
  canvas->format = NULL;
  canvas->tile_width  = TILE_SIZE;
  canvas->tile_height = TILE_SIZE;

  canvas->cache = tile_cache_new ();
  tile_cache_set_size (canvas->cache, CANVAS_CACHE_TILES);
  canvas->system_cache = tile_cache_new ();
  tile_cache_set_wash_percentage (canvas->cache, CANVAS_CACHE_WASH_PERCENTAGE);
  tile_cache_set_size (canvas->system_cache, CANVAS_SYSTEM_CACHE_TILES);
  canvas->n_dirty = 0;
  //canvas_set_path (canvas, CANVAS_PATH);
}

GType
canvas_get_type (void)
{
  static GType type = 0;

  if (type == 0)
  {
    static const GTypeInfo info =
    {
      sizeof (CanvasClass),
      (GBaseInitFunc)     NULL,
      (GBaseFinalizeFunc) NULL,
      (GClassInitFunc)    canvas_class_init,
      NULL,               /* class_finalize */
      NULL,               /* class_data */
      sizeof (Canvas),
      0,                  /* n_preallocs */
      (GInstanceInitFunc) canvas_init
    };

    type = g_type_register_static (G_TYPE_OBJECT,
                                   "Canvas", &info, 0);
  }

  return type;
}

Canvas *canvas_new ()
{
  Canvas *canvas = g_object_new (canvas_get_type(), NULL);

  return canvas;
}

void
canvas_set_undo_stack (Canvas    *canvas,
                       UndoStack *undo_stack)
{
  canvas->undo_stack = undo_stack;
}

UndoStack *
canvas_get_undo_stack (Canvas *canvas)
{
  return canvas->undo_stack;
}

void
canvas_add_dirty (Canvas *canvas,
                  gint    x,
                  gint    y)
{
  //x = x & 0xffffff;
  //y = y & 0xffffff;
  if (canvas_is_dirty (canvas, x,y))
    return;
  if (canvas->n_dirty+1 >= MAX_DIRTY)
    {
      g_warning ("maximum number of dirty tiles exceeded for canvas %i", canvas->n_dirty);
      return;
    }
  else
    {
      gint c = canvas->n_dirty++;
      canvas->dirty[c][0]=x;
      canvas->dirty[c][1]=y;
    }

    /*  add undo */
    if (canvas->undo_stack       && 
        y < tile_tile (MIDDLE)   &&
        x < tile_tile (MIDDLE))
      {
        Tile * tile = canvas_get_tile (canvas, x, y);
        g_object_ref (G_OBJECT (tile));
        undo_stack_add_tile (canvas->undo_stack, tile);
        g_object_unref (G_OBJECT (tile));
      }
}

void
canvas_flush_dirty (Canvas *canvas)
{
  canvas->n_dirty = 0;
}

static gboolean contains (gint array[][2], gint n, gint x, gint y)
{
  gint i;
  for (i=0; i<n; i++)
    {
      if (array[i][0] == x &&
          array[i][1] == y)
        {
          return TRUE;
        }
    }
  return FALSE;
}

gboolean
canvas_is_dirty (Canvas *canvas,
                 gint    x,
                 gint    y)
{
  //x = x & 0xffffff;
  //y = y & 0xffffff;

  if (canvas->n_dirty+1 >= MAX_DIRTY)
    return TRUE;

  return contains (canvas->dirty, canvas->n_dirty, x, y);
}

static Tile *canvas_get_tile_id   (Canvas      *canvas,
                                   const gchar *id,
                                   gint         x,
                                   gint         y);

static char *
make_id (gint x, gint y)
{
  char buf[64];

  sprintf (buf, "%c/%c/%c/%06X_%06X", hex_digits[x%16], hex_digits[y%16], hex_digits[(x/16)%16], x, y);
  return g_strdup(buf);
}

gboolean coord_is_user (gint x, gint y)
{
  gint tmiddle = tile_tile(MIDDLE);

  while (tmiddle < x && tmiddle < y)
    {
      tmiddle += tmiddle/2;
    }
  return !(x > tmiddle || y > tmiddle);
}



/* returns TRUE if something was done */
gboolean
canvas_idle (Canvas *canvas)
{
  gboolean did_something = FALSE;

  did_something |= tile_cache_wash (canvas->cache);
  did_something |= tile_cache_wash (canvas->system_cache);

  return did_something;
}

gint
canvas_get_tile_width (Canvas *canvas)
{
  return canvas->tile_width;
}

gint
canvas_get_tile_height (Canvas *canvas)
{
  return canvas->tile_height;
}


gint
canvas_store_tile (Canvas *canvas,
                   Tile   *tile)
{
  gchar *id;
 
  g_assert (canvas); 
  if (!canvas || !tile)
    return -1;

  id = make_id (tile_get_x (tile), tile_get_y (tile));

  if (id)
    {
       char buf[128];
       sprintf (buf, "%s/%s.png", canvas->path, id);
       tile_save (tile, buf);
       g_free (id);
    }
  return 0;
}

Tile *
canvas_get_tile (Canvas *canvas,
                 gint    x,
                 gint    y)
{
  Tile  *tile = NULL;

  g_assert (canvas);

  if (y >= tile_tile (ALLOCATOR_Y) &&
      x <= tile_tile (MIDDLE))
    {
      tile = tile_cache_get_tile (canvas->system_cache, x, y);
    }
  else
    {
      tile = tile_cache_get_tile (canvas->cache, x, y);
    }
  

  if (tile)
    return tile;
  
  {
    gchar *id;
      
    id = make_id (x,y);

    if (id)
    {
      tile = canvas_get_tile_id (canvas, id, x, y);
      g_free (id);
    }

  if (tile)
    {
      if (y >= tile_tile(ALLOCATOR_Y) &&
          x <= tile_tile(MIDDLE))
        {
          tile_cache_insert (canvas->system_cache, tile);
          /* transfer ownership to cache */
          g_object_unref (G_OBJECT (tile));
        }
      else
        {
          tile_cache_insert (canvas->cache, tile);
          /* transfer ownership to cache */
          g_object_unref (G_OBJECT (tile));
        }
    }
  }
  return tile;
}


static gint prepare_storage (Canvas *canvas)
{
  if (!g_file_test (canvas->path, G_FILE_TEST_IS_DIR))
    {
      if (g_mkdir (canvas->path, 0777))
        {
          g_error ("failed to created canvas storage dir %s", canvas->path);
          return -1;
        }
    }
  return 0;
}

void
canvas_set_path (Canvas      *canvas,
                 const gchar *path)
{
  if (canvas->path)
    g_free (canvas->path);
  canvas->path = NULL;
  if (path)
    canvas->path = g_strdup (path);
  prepare_storage (canvas);
}

const gchar *
canvas_get_path (Canvas *canvas)
{
  return canvas->path;
}


#include "SDL.h"

static void
blank_tile (Tile *tile)
{
  guchar *data   = tile_get_data (tile);
  gint    width  = tile_get_width (tile);
  gint    height = tile_get_height (tile);

  memset (data, 0xff, width * height *4);
}

/* this is the only place that actually should
 * instantiate tiles, when the cache is large enough
 * that should make sure we don't hit this function
 * too often.
 */
static Tile *canvas_get_tile_id (Canvas      *canvas,
                                 const gchar *id,
                                 gint         x,
                                 gint         y)
{
  Tile *tile;
  char path[512];
  sprintf (path, "%s/%s.png", canvas->path, id);
  tile = tile_load (path);

  if (!tile)
    {
      tile = tile_new (canvas->tile_width,
                       canvas->tile_height,
                       canvas->format);
      blank_tile (tile);
    }

  tile_set_canvas (tile, canvas);
  tile_set_x (tile, x);
  tile_set_y (tile, y);
  return tile;
}


#include "scale.h"

void
canvas_update_zoom (Canvas *canvas)
{
  gint levels = SUBDIVIDE_LEVELS;
  gint i;
  gint zoom[MAX_DIRTY][2];
  gint n_zoom = 0;
  gint dirty[MAX_DIRTY][2];
  gint n_dirty = 0;

  for (i=0; i<canvas->n_dirty; i++)
    {
      gint x = canvas->dirty[i][0];
      gint y = canvas->dirty[i][1];

      /* do not create scaled down version of system region of canvas */
      if (y>=MIDDLE)
        continue;

      gint zx = (tile_tile(TILE_ZOOM_START)*2 + x)/2;
      gint zy = (tile_tile(TILE_ZOOM_START)*2 + y)/2;
      
      if (!contains (zoom, n_zoom, zx, zy))
        {
          zoom[n_zoom][0]=zx;
          zoom[n_zoom++][1]=zy;
        }
    }

  {
    gint i;
    for (i=0;i<canvas->n_dirty;i++)
      {
        dirty[i][0]=canvas->dirty[i][0];
        dirty[i][1]=canvas->dirty[i][1];
      }
    n_dirty = canvas->n_dirty;
  }

  while(levels--)
    {
      gint zoomB[MAX_DIRTY][2];
      gint n_zoomB = 0;

      // fprintf (stderr, "%i %i\n", levels, n_zoom);

      for (i=0; i<n_zoom; i++)
        {
          Tile *dst;
         
          Tile *src[4]={NULL,NULL,NULL,NULL};
          gint x,y;
          gint sx,sy;
         
          x = zoom[i][0]; 
          y = zoom[i][1];

          sx = x * 2 - tile_tile(TILE_ZOOM_START)*2;
          sy = y * 2 - tile_tile(TILE_ZOOM_START)*2;

          dst = canvas_get_tile (canvas, x,y);
          g_object_ref (G_OBJECT (dst));

          if (contains (dirty, n_dirty, sx+0, sy+0))
           {
             src[0] = canvas_get_tile (canvas, sx+0, sy+0);
             g_object_ref (G_OBJECT (src[0]));
           }
          if (contains (dirty, n_dirty, sx+1, sy+0))
           {
             src[1] = canvas_get_tile (canvas, sx+1, sy+0);
             g_object_ref (G_OBJECT (src[1]));
           }
          if (contains (dirty, n_dirty, sx+0, sy+1))
           {
             src[2] = canvas_get_tile (canvas, sx+0, sy+1);
             g_object_ref (G_OBJECT (src[2]));
           }
          if (contains (dirty, n_dirty, sx+1, sy+1))
           {
             src[3] = canvas_get_tile (canvas, sx+1, sy+1);
             g_object_ref (G_OBJECT (src[3]));
           }

          scale (src, dst);
          {
            gint i;
            for (i=0; i<4; i++)
              if (src[i])
                g_object_unref (G_OBJECT (src[i]));
          }
          g_object_unref (G_OBJECT (dst));

          {
            gint zx = (tile_tile(TILE_ZOOM_START)*2 + x)/2;
            gint zy = (tile_tile(TILE_ZOOM_START)*2 + y)/2;

            if (!contains(zoomB, n_zoomB, zx,zy))
              {
                zoomB[n_zoomB][0]=zx;
                zoomB[n_zoomB][1]=zy;
                n_zoomB++;
              }
          }
        }
      {
        gint i;
        for (i=0;i<n_zoom;i++)
          {
            dirty[i][0]=zoom[i][0];
            dirty[i][1]=zoom[i][1];
          }
        n_dirty = n_zoom;
      }
      {
        gint i;
        for (i=0;i<n_zoomB;i++)
          {
            zoom[i][0]=zoomB[i][0];
            zoom[i][1]=zoomB[i][1];
          }
        n_zoom = n_zoomB;
      }
    }
}


static gboolean
canvas_has_tile (Canvas *canvas,
                 gint x,
                 gint y)
{
  Tile  *tile = NULL;

  if (y >= tile_tile (ALLOCATOR_Y) &&
      x <= tile_tile (MIDDLE))
    {
      tile = tile_cache_get_tile (canvas->system_cache, x, y);
    }
  else
    {
      tile = tile_cache_get_tile (canvas->cache, x, y);
    }
  return tile!=NULL;
}

/*
 * Prefetch a region of the canvas:
 * params: 
 *   x0     upperleft horizontal coordinate in canvas pixel units
 *   y0     upperleft vertical coordinate in canvas pixel units
 *   width  width in canvas pixel units
 *   height height in canvas pixel units
 *   scale  the scale level a view would be set to to render this
 *          region. (set to 65536 for 1:1) 
 *
 * returns true if a tile was fetched.
 */
gboolean
canvas_prefetch (Canvas    *canvas,
                 gint       x0,
                 gint       y0,
                 gint       width,
                 gint       height,
                 gint       scale,
                 gboolean   do_context)
{
  gint  x=0,y;
  gint  sscale;
  glong sx, sy;
  gint  scaled_tile_width;
  gint  scaled_tile_height;
  gint  context_tiles = do_context?1:0;

  sx = x0;
  sy = y0;
  sscale = scale;

  {
    while (sscale < 65536 * MIN_TILE_SCALE)
      {
        sx = (sx + (TILE_ZOOM_START*2))/2;
        sy = (sy + (TILE_ZOOM_START*2))/2;
        sscale = sscale * 2;
      }
  }

  scaled_tile_width  = canvas->tile_width  * sscale / 65536;
  scaled_tile_height = canvas->tile_height * sscale / 65536;

  for (y=-context_tiles;
       (y-1-context_tiles) * scaled_tile_height < height * scale / 65536;
       y++)
         for (x=-context_tiles;
              (x-1-context_tiles) * scaled_tile_width < width * scale / 65536;
              x++)
           {
             Tile *tile;
             gint tile_x = (tile_tile (sx) + x) & 0xffffff;
             gint tile_y = (tile_tile (sy) + y) & 0xffffff;

             if (!canvas_has_tile (canvas, tile_x, tile_y))
               {
                 tile = canvas_get_tile (canvas, tile_x, tile_y);
                 return TRUE;
               }
           }
  return FALSE;
}



void
canvas_pick (Canvas *canvas,
             gint    x,
             gint    y,
             guint  *r,
             guint  *g,
             guint  *b)
{
  Tile *tile;

  gint tile_width = canvas_get_tile_width (canvas);
  gint tile_height = canvas_get_tile_height (canvas);
  
  gint tile_x = x / tile_width;
  gint tile_y = y / tile_height;

  if (x<0)
    tile_x -=1;
  if (y<0)
    tile_y -=1;

    {
      gint tx,ty;
      guchar *data;

      tx = x % tile_width;
      ty = y % tile_height;

      if (x<0)
        tx += tile_width;
      if (y<0)
        ty += tile_height;

      /* note: we are not referencing the tile, this should nevertheless work
       * since we assume a single thread of operation and that the canvas will
       * have a cached entry for this tile */
      tile = canvas_get_tile (canvas, tile_x, tile_y);

      data = tile_get_data (tile) + tile_width * ty * 4 + tx * 4;
      *r = data[0] * 256;
      *g = data[1] * 256;
      *b = data[2] * 256;
    }
}



void
canvas_outline_box (Canvas *canvas,
                    gint    x0,
                    gint    y0,
                    gint    width,
                    gint    height,
                    gint    red,
                    gint    green,
                    gint    blue,
                    gint    alpha)
{
  canvas_fill_rect (canvas, x0, y0, width, 1, red, green, blue, alpha);
  canvas_fill_rect (canvas, x0, y0 + height, width, 1, red, green, blue, alpha);
  canvas_fill_rect (canvas, x0, y0, 1, height, red, green, blue, alpha);
  canvas_fill_rect (canvas, x0 + width, y0, 1, height, red, green, blue, alpha);
}


void
canvas_fill_rect (Canvas *canvas,
                  gint    x,
                  gint    y,
                  gint    width,
                  gint    height,
                  gint    red,
                  gint    green,
                  gint    blue,
                  gint    alpha)
{
  guchar r = red*255/65536;
  guchar g = green*255/65536;
  guchar b = blue*255/65536;
  guchar a = alpha*255/65536;
  guchar *buf = g_malloc (4 * width * height);
  if (buf)
    {
      CanvasRect *rect = canvas_rect_new (canvas, x, y, width, height);
      gint i;
      for (i=0;i<width*height;i++)
        {
          buf[i*4 + 0] = r;
          buf[i*4 + 1] = g;
          buf[i*4 + 2] = b;
          buf[i*4 + 3] = a;
        }
      canvas_rect_import_buf (rect, buf);
      g_object_unref (rect); 
      g_free (buf);
    }
}

