/*  horizon, a sketch, note and scribbling application.
    Copyright (C) 2005, 2006 Øyvind Kolås <pippin@gimp.org>
                             Eero Tanskanen <yendor@nic.fi>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
*/
#include <glib.h>
#include <math.h>
#include "tool_text.h"
#include "tool.h"
#include "brush.h"
#include "toolkit/slider.h"
#include "toolkit/view.h"
#include "toolkit/navigator.h"
#include "canvas/canvas.h"
#include "teleport.h"
#include "actions.h"
#include "link.h"

#include "config.h"

struct _ToolText {
  Tool    tool;
  View   *thumbnail;

  gint    left_margin;
  gint    x;
  gint    y;
  gint    base_line;
  gint    font_size;

  gint    char_width[256];
  gint    chars;
};


static void
tool_text_start (Tool  *tool,
                gint    x,
                gint    y,
                gint    pressure,
                gulong  time)
{
  Easel *easel = tool->easel;
  View  *view  = (View*)easel;

  x /= SPP;
  y /= SPP;
 
  view_set_rect_x (view, x);
  view_set_rect_y (view, y);

  view_set_rect_width (view, 200);
  view_set_rect_height (view, 200);
  
  tool->prev_x = x;  /* prev is used as the /grabbed coordinate/ in pan tool */
  tool->prev_y = y;
  tool->prev_time = time;
}

static void
tool_text_continue (Tool  *tool,
                   gint   x,
                   gint   y,
                   gint   pressure,
                   gulong time)
{
  ToolText *tool_text = (ToolText*) tool;
  Easel *easel = tool->easel;
  View  *view  = (View*)easel;
  x /= SPP;
  y /= SPP;

  view_set_rect_width  (view, x-tool->prev_x);
  view_set_rect_height (view, y-tool->prev_y);

  tool_text->left_margin = tool->prev_x * SPP;
  tool_text->base_line = tool->prev_y * SPP + (y-tool->prev_y) * SPP;
  tool_text->font_size = (y-tool->prev_y);
  tool_text->x = tool_text->left_margin;
  tool_text->y = tool_text->base_line;
  tool_text->chars=0;
  
  tool->prev_pressure=pressure;
  tool->prev_time=time;
}

void
tool_text_process_coordinate (Tool   *tool,
                             gint    x,
                             gint    y,
                             gint    pressure,
                             gulong  time)
{
  if (time - tool->prev_time > 200)
    {
      tool_text_start (tool, x, y, pressure, time);
    }
  else
    {
      tool_text_continue (tool, x, y, pressure, time);
    }
}

int
tool_text_flush (Tool *tool)
{
  return 0;
}

#include "vectors/font.h"
#include "vectors/path.h"

void tool_text_feed (Tool *tool,
                     const gchar *buf)
{
  ToolText *tool_text = (ToolText*) tool;

  if (buf && buf[0]==13)
    {
      tool_text->x = tool_text->left_margin;
      tool_text->y += tool_text->font_size * SPP * 1.5;
      tool_text->chars=0;
    }
  else if (buf && buf[0]==8)
    {
      if (tool_text->chars>0)
        {
          undo_stack_undo (canvas_get_undo_stack (tool->canvas));
          tool_text->x = tool_text->char_width[--tool_text->chars];
        }
    }
  else if (buf && buf[0])
    {
      HorizonFont *font = font_new ();
      Path *path;
      font_set_size (font, tool_text->font_size);
      path = font_path (font, NULL, tool_text->x, tool_text->y, 1000, buf);

      path_set_line_width (path, 65536 * tool_text->font_size * 0.001);
      path_set_hardness (path, 65536 * 0.75);

      undo_stack_new_step (canvas_get_undo_stack (tool->canvas));
      
      path_stroke (path, tool->canvas);
      tool_text->char_width[tool_text->chars++] = tool_text->x;
      tool_text->x = path_last_x (path);
      
      path_destroy (path);
      font_destroy (font);
    }
}

void tool_text_fill_tool_specific_settings_box (Tool *tool)
{
  ToolText *tool_text = (ToolText*) tool;
  Box *vbox;
  
  vbox = vbox_new (1);

  tool->tool_settings_box = vbox;
  
  /* Create thumbnail view. */
  tool_text->thumbnail = (View*)navigator_new (tool->canvas);
  navigator_set_easel ((Navigator*)tool_text->thumbnail, tool->easel);
  box_set_child (vbox, 0, (Box*)tool_text->thumbnail);

  tool_text_flush (tool);
}

static void selected (Tool *tool)
{
  Easel *easel = tool->easel;
  View  *view  = (View*)easel;

  gint color[4]         = {0,     0,     0,     65536*0.1};
  gint border_color[4]  = {65535, 65535, 65536, 65535};
  gint surround_color[4]= {65535, 65535, 65536, 0};

  view_set_rect_color (view, color);
  view_set_rect_border_color (view, border_color);
  view_set_rect_surround_color (view, surround_color);
}

static void
deselected (Tool *tool)
{
  Easel *easel = tool->easel;
  View  *view = (View*)easel;

  view_set_rect_visible (view, FALSE);
}


static void change_mode (Tool *tool)
{
  horizon_show_prefs (tool->horizon);
}

Tool *
tool_text_new ()
{
  ToolText *tool_text = g_malloc0 (sizeof (ToolText));
  Tool *tool = (Tool*) tool_text;
  
  tool_init (tool);
  
  tool->process_coordinate = tool_text_process_coordinate;
  tool->flush = tool_text_flush;
  tool->tool_fill_tool_specific_settings_box = tool_text_fill_tool_specific_settings_box;
  tool->short_name = "Text";
  tool->change_mode = change_mode;
  tool->deselect = deselected;
  tool->select = selected;
 
  return tool;
}

void
tool_text_destroy (Tool *tool)
{
  g_free ((ToolText *)tool);
}
